@@ -1,5 +1,3 @@
-# coding: utf-8
-
 import inspect
 import os
 import re
@@ -7,12 +5,14 @@ import warnings
 from io import BufferedReader
 from io import BytesIO
 from typing import Any
+from typing import List
 from typing import Optional
 from unittest.mock import Mock
 from unittest.mock import patch
 
 import pytest
 import requests
+import urllib3
 from requests.exceptions import ChunkedEncodingError
 from requests.exceptions import ConnectionError
 from requests.exceptions import HTTPError
@@ -21,6 +21,7 @@ from urllib3.util.retry import Retry
 
 import responses
 from responses import BaseResponse
+from responses import Call
 from responses import CallbackResponse
 from responses import PassthroughResponse
 from responses import Response
@@ -122,7 +123,7 @@ def test_response_with_instance():
         ),
     ],
 )
-def test_replace(original, replacement):
+def test_replace(original, replacement):  # type: ignore[misc]
     @responses.activate
     def run():
         responses.add(responses.GET, "http://example.com/one", body="test1")
@@ -156,7 +157,7 @@ def test_replace(original, replacement):
         (re.compile(r"http://example\.com/one"), "http://example.com/one"),
     ],
 )
-def test_replace_error(original, replacement):
+def test_replace_error(original, replacement):  # type: ignore[misc]
     @responses.activate
     def run():
         responses.add(responses.GET, original)
@@ -202,7 +203,7 @@ def test_replace_response_object_error():
         ),
     ],
 )
-def test_upsert_replace(original, replacement):
+def test_upsert_replace(original, replacement):  # type: ignore[misc]
     @responses.activate
     def run():
         responses.add(responses.GET, "http://example.com/one", body="test1")
@@ -240,7 +241,7 @@ def test_upsert_replace(original, replacement):
         ),
     ],
 )
-def test_upsert_add(original, replacement):
+def test_upsert_add(original, replacement):  # type: ignore[misc]
     @responses.activate
     def run():
         responses.add(responses.GET, "http://example.com/one", body="test1")
@@ -298,7 +299,7 @@ def test_remove():
         ),
     ],
 )
-def test_response_equality(args1, kwargs1, args2, kwargs2, expected):
+def test_response_equality(args1, kwargs1, args2, kwargs2, expected):  # type: ignore[misc]
     o1 = BaseResponse(*args1, **kwargs1)
     o2 = BaseResponse(*args2, **kwargs2)
     assert (o1 == o2) is expected
@@ -801,34 +802,70 @@ def test_base_response_get_response():
         resp.get_response(requests.PreparedRequest())
 
 
-def test_custom_adapter():
-    @responses.activate
-    def run():
-        url = "http://example.com"
-        responses.add(responses.GET, url, body=b"test")
+class TestAdapters:
+    class CustomAdapter(requests.adapters.HTTPAdapter):
+        """Classic custom adapter."""
 
-        calls = [0]
+        def send(self, *a, **k):
+            return super().send(*a, **k)
 
-        class DummyAdapter(requests.adapters.HTTPAdapter):
-            def send(self, *a, **k):
-                calls[0] += 1
-                return super().send(*a, **k)
+    class PositionalArgsAdapter(requests.adapters.HTTPAdapter):
+        """Custom adapter that sends only positional args.
+        See https://github.com/getsentry/responses/issues/642 for more into.
+        """
 
-        # Test that the adapter is actually used
-        session = requests.Session()
-        session.mount("http://", DummyAdapter())
+        def send(
+            self,
+            request,
+            stream=False,
+            timeout=None,
+            verify=True,
+            cert=None,
+            proxies=None,
+        ):
+            return super().send(request, stream, timeout, verify, cert, proxies)
 
-        session.get(url, allow_redirects=False)
-        assert calls[0] == 1
+    class PositionalArgsIncompleteAdapter(requests.adapters.HTTPAdapter):
+        """Custom adapter that sends only positional args.
+        Not all arguments are forwarded to the send method.
+                    See https://github.com/getsentry/responses/issues/642 for more into.
+        """
 
-        # Test that the response is still correctly emulated
-        session = requests.Session()
-        session.mount("http://", DummyAdapter())
+        def send(
+            self,
+            request,
+            stream=False,
+            timeout=None,
+            verify=True,
+            # following args are intentionally not forwarded
+            cert=None,
+            proxies=None,
+        ):
+            return super().send(request, stream, timeout, verify)
 
-        resp = session.get(url)
-        assert_response(resp, "test")
+    @pytest.mark.parametrize(
+        "adapter_class",
+        (CustomAdapter, PositionalArgsAdapter, PositionalArgsIncompleteAdapter),
+    )
+    def test_custom_adapter(self, adapter_class):  # type: ignore[misc]
+        """Test basic adapter implementation and that responses can patch them properly."""
 
-    run()
+        @responses.activate
+        def run():
+            url = "http://example.com"
+            responses.add(responses.GET, url, body=b"test adapter")
+
+            # Test that the adapter is actually used
+            session = requests.Session()
+            adapter = adapter_class()
+            session.mount("http://", adapter)
+            with patch.object(adapter, "send", side_effect=adapter.send) as mock_send:
+                resp = session.get(url, allow_redirects=False)
+
+            assert mock_send.call_count == 1
+            assert_response(resp, "test adapter")
+
+        run()
 
 
 def test_responses_as_context_manager():
@@ -865,17 +902,17 @@ def test_activate_doesnt_change_signature():
 
 
 @pytest.fixture
-def my_fruit():
+def my_fruit():  # type: ignore[misc]
     return "apple"
 
 
 @pytest.fixture
-def fruit_basket(my_fruit):
+def fruit_basket(my_fruit):  # type: ignore[misc]
     return ["banana", my_fruit]
 
 
 @pytest.mark.usefixtures("my_fruit", "fruit_basket")
-class TestFixtures(object):
+class TestFixtures:
     """
     Test that pytest fixtures work well with 'activate' decorator
     """
@@ -889,7 +926,7 @@ class TestFixtures(object):
 
 def test_activate_mock_interaction():
     @patch("sys.stdout")
-    def test_function(mock_stdout):
+    def test_function(mock_stdout):  # type: ignore[misc]
         return mock_stdout
 
     decorated_test_function = responses.activate(test_function)
@@ -923,7 +960,7 @@ def test_activate_doesnt_change_signature_with_return_type():
 
 
 def test_activate_doesnt_change_signature_for_method():
-    class TestCase(object):
+    class TestCase:
         def test_function(self, a, b=None):
             return self, a, b
 
@@ -1008,7 +1045,7 @@ def test_response_cookies_multiple():
 
 @pytest.mark.parametrize("request_stream", (True, False, None))
 @pytest.mark.parametrize("responses_stream", (True, False, None))
-def test_response_cookies_session(request_stream, responses_stream):
+def test_response_cookies_session(request_stream, responses_stream):  # type: ignore[misc]
     @responses.activate
     def run():
         url = "https://example.com/path"
@@ -1063,10 +1100,10 @@ def test_response_filebody():
     def run():
         current_file = os.path.abspath(__file__)
         with responses.RequestsMock() as m:
-            with open(current_file, "r", encoding="utf-8") as out:
+            with open(current_file, encoding="utf-8") as out:
                 m.add(responses.GET, "http://example.com", body=out.read(), stream=True)
                 resp = requests.get("http://example.com", stream=True)
-            with open(current_file, "r", encoding="utf-8") as out:
+            with open(current_file, encoding="utf-8") as out:
                 assert resp.text == out.read()
 
     run()
@@ -1158,7 +1195,7 @@ def test_assert_all_requests_fired_multiple():
 def test_allow_redirects_samehost():
     redirecting_url = "http://example.com"
     final_url_path = "/1"
-    final_url = "{0}{1}".format(redirecting_url, final_url_path)
+    final_url = f"{redirecting_url}{final_url_path}"
     url_re = re.compile(r"^http://example.com(/)?(\d+)?$")
 
     def request_callback(request):
@@ -1172,7 +1209,7 @@ def test_allow_redirects_samehost():
                 n = 1
             else:
                 n = 0
-            redirect_headers = {"location": "/{0!s}".format(n)}
+            redirect_headers = {"location": f"/{n!s}"}
             return 301, redirect_headers, None
 
     def run():
@@ -1295,6 +1332,29 @@ def test_headers():
     assert_reset()
 
 
+def test_headers_deduplicated_content_type():
+    """Test to ensure that we do not have two values for `content-type`.
+
+    For more details see https://github.com/getsentry/responses/issues/644
+    """
+
+    @responses.activate
+    def run():
+        responses.get(
+            "https://example.org/",
+            json={},
+            headers={"Content-Type": "application/json"},
+        )
+        responses.start()
+
+        resp = requests.get("https://example.org/")
+
+        assert resp.headers["Content-Type"] == "application/json"
+
+    run()
+    assert_reset()
+
+
 def test_content_length_error(monkeypatch):
     """
     Currently 'requests' does not enforce content length validation,
@@ -1324,15 +1384,13 @@ def test_content_length_error(monkeypatch):
     # Type errors here and on 1250 are ignored because the stubs for requests
     # are off https://github.com/python/typeshed/blob/f8501d33c737482a829c6db557a0be26895c5941
     #   /stubs/requests/requests/packages/__init__.pyi#L1
-    original_init = getattr(requests.packages.urllib3.HTTPResponse, "__init__")  # type: ignore
+    original_init = getattr(urllib3.HTTPResponse, "__init__")
 
     def patched_init(self, *args, **kwargs):
         kwargs["enforce_content_length"] = True
         original_init(self, *args, **kwargs)
 
-    monkeypatch.setattr(
-        requests.packages.urllib3.HTTPResponse, "__init__", patched_init  # type: ignore
-    )
+    monkeypatch.setattr(urllib3.HTTPResponse, "__init__", patched_init)
 
     run()
     assert_reset()
@@ -1498,9 +1556,17 @@ def test_auto_calculate_content_length_doesnt_override_existing_value():
             headers={"Content-Length": "2"},
             auto_calculate_content_length=True,
         )
-        resp = requests.get(url)
-        assert_response(resp, "test")
-        assert resp.headers["Content-Length"] == "2"
+
+        if urllib3.__version__ < "2":
+            resp = requests.get(url)
+            assert_response(resp, "test")
+            assert resp.headers["Content-Length"] == "2"
+        else:
+            with pytest.raises(ChunkedEncodingError) as excinfo:
+                requests.get(url)
+            assert "IncompleteRead(4 bytes read, -2 more expected)" in str(
+                excinfo.value
+            )
 
     run()
     assert_reset()
@@ -1587,196 +1653,241 @@ def test_multiple_methods():
     assert_reset()
 
 
-def test_passthrough_flag(httpserver):
-    httpserver.expect_request("/").respond_with_data("OK", content_type="text/plain")
-    url = httpserver.url_for("/")
-
-    response = Response(responses.GET, url, body="MOCK")
-
-    @responses.activate
-    def run_passthrough():
-        responses.add(response)
-        resp = requests.get(url)
-        assert_response(resp, "OK")
-
-    @responses.activate
-    def run_mocked():
-        responses.add(response)
-        resp = requests.get(url)
-        assert_response(resp, "MOCK")
-
-    run_mocked()
-    assert_reset()
-
-    response.passthrough = True
-    run_passthrough()
-    assert_reset()
-
-
-def test_passthrough_kwarg(httpserver):
-    httpserver.expect_request("/").respond_with_data("OK", content_type="text/plain")
-    url = httpserver.url_for("/")
-
-    def configure_response(passthrough):
-        responses.get(url, body="MOCK", passthrough=passthrough)
-
-    @responses.activate
-    def run_passthrough():
-        configure_response(passthrough=True)
-        resp = requests.get(url)
-        assert_response(resp, "OK")
-
-    @responses.activate
-    def run_mocked():
-        configure_response(passthrough=False)
-        resp = requests.get(url)
-        assert_response(resp, "MOCK")
-
-    run_mocked()
-    assert_reset()
-
-    run_passthrough()
-    assert_reset()
-
-
-def test_passthrough_response(httpserver):
-    httpserver.expect_request("/").respond_with_data("OK", content_type="text/plain")
-    url = httpserver.url_for("/")
-
-    @responses.activate
-    def run():
-        responses.add(PassthroughResponse(responses.GET, url))
-        responses.add(responses.GET, "{}/one".format(url), body="one")
-        responses.add(responses.GET, "http://example.com/two", body="two")
-
-        resp = requests.get("http://example.com/two")
-        assert_response(resp, "two")
-        resp = requests.get("{}/one".format(url))
-        assert_response(resp, "one")
-        resp = requests.get(url)
-        assert_response(resp, "OK")
-
-        assert len(responses.calls) == 3
-        responses.assert_call_count(url, 1)
-
-    run()
-    assert_reset()
-
-
-def test_passthrough_response_stream(httpserver):
-    httpserver.expect_request("/").respond_with_data("OK", content_type="text/plain")
-
-    @responses.activate
-    def run():
+class TestPassthru:
+    def test_passthrough_flag(self, httpserver):
+        httpserver.expect_request("/").respond_with_data(
+            "OK", content_type="text/plain"
+        )
         url = httpserver.url_for("/")
-        responses.add(PassthroughResponse(responses.GET, url))
-        content_1 = requests.get(url).content
-        with requests.get(url, stream=True) as resp:
-            content_2 = resp.raw.read()
-        assert content_1 == content_2
 
-    run()
-    assert_reset()
+        response = Response(responses.GET, url, body="MOCK")
 
-
-def test_passthru_prefixes(httpserver):
-    httpserver.expect_request("/").respond_with_data("OK", content_type="text/plain")
-    url = httpserver.url_for("/")
-
-    @responses.activate
-    def run_constructor_argument():
-        with responses.RequestsMock(passthru_prefixes=(url,)):
+        @responses.activate
+        def run_passthrough():
+            responses.add(response)
             resp = requests.get(url)
             assert_response(resp, "OK")
 
-    @responses.activate
-    def run_property_setter():
-        with responses.RequestsMock() as m:
-            m.passthru_prefixes = tuple([url])
+        @responses.activate
+        def run_mocked():
+            responses.add(response)
+            resp = requests.get(url)
+            assert_response(resp, "MOCK")
+
+        run_mocked()
+        assert_reset()
+
+        response.passthrough = True
+        run_passthrough()
+        assert_reset()
+
+    def test_passthrough_kwarg(self, httpserver):
+        httpserver.expect_request("/").respond_with_data(
+            "OK", content_type="text/plain"
+        )
+        url = httpserver.url_for("/")
+
+        def configure_response(passthrough):
+            responses.get(url, body="MOCK", passthrough=passthrough)
+
+        @responses.activate
+        def run_passthrough():
+            configure_response(passthrough=True)
             resp = requests.get(url)
             assert_response(resp, "OK")
 
-    run_constructor_argument()
-    assert_reset()
-    run_property_setter()
-    assert_reset()
+        @responses.activate
+        def run_mocked():
+            configure_response(passthrough=False)
+            resp = requests.get(url)
+            assert_response(resp, "MOCK")
 
+        run_mocked()
+        assert_reset()
 
-def test_passthru(httpserver):
-    httpserver.expect_request("/").respond_with_data("OK", content_type="text/plain")
-    url = httpserver.url_for("/")
+        run_passthrough()
+        assert_reset()
 
-    @responses.activate
-    def run():
-        responses.add_passthru(url)
-        responses.add(responses.GET, "{}/one".format(url), body="one")
-        responses.add(responses.GET, "http://example.com/two", body="two")
+    def test_passthrough_response(self, httpserver):
+        httpserver.expect_request("/").respond_with_data(
+            "OK", content_type="text/plain"
+        )
+        url = httpserver.url_for("/")
 
-        resp = requests.get("http://example.com/two")
-        assert_response(resp, "two")
-        resp = requests.get("{}/one".format(url))
-        assert_response(resp, "one")
-        resp = requests.get(url)
-        assert_response(resp, "OK")
+        @responses.activate
+        def run():
+            responses.add(PassthroughResponse(responses.GET, url))
+            responses.add(responses.GET, f"{url}/one", body="one")
+            responses.add(responses.GET, "http://example.com/two", body="two")
 
-    run()
-    assert_reset()
+            resp = requests.get("http://example.com/two")
+            assert_response(resp, "two")
+            resp = requests.get(f"{url}/one")
+            assert_response(resp, "one")
+            resp = requests.get(url)
+            assert_response(resp, "OK")
 
+            assert len(responses.calls) == 3
+            responses.assert_call_count(url, 1)
 
-def test_passthru_regex(httpserver):
-    httpserver.expect_request(re.compile("^/\\w+")).respond_with_data(
-        "OK", content_type="text/plain"
-    )
-    url = httpserver.url_for("/")
+        run()
+        assert_reset()
 
-    @responses.activate
-    def run():
-        responses.add_passthru(re.compile(f"{url}/\\w+"))
-        responses.add(responses.GET, "{}/one".format(url), body="one")
-        responses.add(responses.GET, "http://example.com/two", body="two")
+    def test_passthrough_response_stream(self, httpserver):
+        httpserver.expect_request("/").respond_with_data(
+            "OK", content_type="text/plain"
+        )
 
-        resp = requests.get("http://example.com/two")
-        assert_response(resp, "two")
-        resp = requests.get(f"{url}/one")
-        assert_response(resp, "one")
-        resp = requests.get(f"{url}/two")
-        assert_response(resp, "OK")
-        resp = requests.get(f"{url}/three")
-        assert_response(resp, "OK")
+        @responses.activate
+        def run():
+            url = httpserver.url_for("/")
+            responses.add(PassthroughResponse(responses.GET, url))
+            content_1 = requests.get(url).content
+            with requests.get(url, stream=True) as resp:
+                content_2 = resp.raw.read()
+            assert content_1 == content_2
 
-    run()
-    assert_reset()
+        run()
+        assert_reset()
 
+    def test_passthru_prefixes(self, httpserver):
+        httpserver.expect_request("/").respond_with_data(
+            "OK", content_type="text/plain"
+        )
+        url = httpserver.url_for("/")
 
-def test_passthru_does_not_persist_across_tests(httpserver):
-    """
-    passthru should be erased on exit from context manager
-    see:
-    https://github.com/getsentry/responses/issues/322
-    """
-    httpserver.expect_request("/").respond_with_data("OK", content_type="text/plain")
+        @responses.activate
+        def run_constructor_argument():
+            with responses.RequestsMock(passthru_prefixes=(url,)):
+                resp = requests.get(url)
+                assert_response(resp, "OK")
 
-    @responses.activate
-    def with_a_passthru():
-        assert not responses.mock.passthru_prefixes
-        responses.add_passthru(re.compile(".*"))
+        @responses.activate
+        def run_property_setter():
+            with responses.RequestsMock() as m:
+                m.passthru_prefixes = tuple([url])
+                resp = requests.get(url)
+                assert_response(resp, "OK")
 
-        # wrap request that is passed through with another mock. That helps
-        # to avoid issues if real URL is unavailable, allow to run tests offline
-        with responses.RequestsMock(target="responses._real_send") as rsp:
-            rsp.add(responses.GET, "https://example66.ru", status=969)
-            response = requests.get("https://example66.ru")
+        run_constructor_argument()
+        assert_reset()
+        run_property_setter()
+        assert_reset()
 
+    def test_passthru(self, httpserver):
+        httpserver.expect_request("/").respond_with_data(
+            "OK", content_type="text/plain"
+        )
+        url = httpserver.url_for("/")
+
+        @responses.activate
+        def run():
+            responses.add_passthru(url)
+            responses.add(responses.GET, f"{url}/one", body="one")
+            responses.add(responses.GET, "http://example.com/two", body="two")
+
+            resp = requests.get("http://example.com/two")
+            assert_response(resp, "two")
+            resp = requests.get(f"{url}/one")
+            assert_response(resp, "one")
+            resp = requests.get(url)
+            assert_response(resp, "OK")
+
+        run()
+        assert_reset()
+
+    def test_passthru_regex(self, httpserver):
+        httpserver.expect_request(re.compile("^/\\w+")).respond_with_data(
+            "OK", content_type="text/plain"
+        )
+        url = httpserver.url_for("/")
+
+        @responses.activate
+        def run():
+            responses.add_passthru(re.compile(f"{url}/\\w+"))
+            responses.add(responses.GET, f"{url}/one", body="one")
+            responses.add(responses.GET, "http://example.com/two", body="two")
+
+            resp = requests.get("http://example.com/two")
+            assert_response(resp, "two")
+            resp = requests.get(f"{url}/one")
+            assert_response(resp, "one")
+            resp = requests.get(f"{url}/two")
+            assert_response(resp, "OK")
+            resp = requests.get(f"{url}/three")
+            assert_response(resp, "OK")
+
+        run()
+        assert_reset()
+
+    def test_passthru_does_not_persist_across_tests(self, httpserver):
+        """
+        passthru should be erased on exit from context manager
+        see:
+        https://github.com/getsentry/responses/issues/322
+        """
+        httpserver.expect_request("/").respond_with_data(
+            "mocked server", status=969, content_type="text/plain"
+        )
+
+        @responses.activate
+        def with_a_passthru():
+            assert not responses.mock.passthru_prefixes
+            responses.add_passthru(re.compile(".*"))
+
+            url = httpserver.url_for("/")
+            response = requests.get(url)
             assert response.status_code == 969
+            assert response.text == "mocked server"
 
-    @responses.activate
-    def without_a_passthru():
-        assert not responses.mock.passthru_prefixes
-        with pytest.raises(requests.exceptions.ConnectionError):
-            requests.get("https://example.com")
+        @responses.activate
+        def without_a_passthru():
+            assert not responses.mock.passthru_prefixes
+            with pytest.raises(requests.exceptions.ConnectionError):
+                requests.get("https://example.com")
 
-    with_a_passthru()
-    without_a_passthru()
+        with_a_passthru()
+        without_a_passthru()
+
+    def test_passthru_unicode(self):
+        @responses.activate
+        def run():
+            with responses.RequestsMock() as m:
+                url = "http://موقع.وزارة-الاتصالات.مصر/"
+                clean_url = "http://xn--4gbrim.xn----ymcbaaajlc6dj7bxne2c.xn--wgbh1c/"
+                m.add_passthru(url)
+                assert m.passthru_prefixes[0] == clean_url
+
+        run()
+        assert_reset()
+
+    def test_real_send_argument(self):
+        def run():
+            # the following mock will serve to catch the real send request from another mock and
+            # will "donate" `unbound_on_send` method
+            mock_to_catch_real_send = responses.RequestsMock(
+                assert_all_requests_are_fired=True
+            )
+            mock_to_catch_real_send.post(
+                "http://send-this-request-through.com", status=500
+            )
+
+            with responses.RequestsMock(
+                assert_all_requests_are_fired=True,
+                real_adapter_send=mock_to_catch_real_send.unbound_on_send(),
+            ) as r_mock:
+                r_mock.add_passthru("http://send-this-request-through.com")
+
+                r_mock.add(responses.POST, "https://example.org", status=200)
+
+                response = requests.post("https://example.org")
+                assert response.status_code == 200
+
+                response = requests.post("http://send-this-request-through.com")
+                assert response.status_code == 500
+
+        run()
+        assert_reset()
 
 
 def test_method_named_param():
@@ -1790,19 +1901,6 @@ def test_method_named_param():
     assert_reset()
 
 
-def test_passthru_unicode():
-    @responses.activate
-    def run():
-        with responses.RequestsMock() as m:
-            url = "http://موقع.وزارة-الاتصالات.مصر/"
-            clean_url = "http://xn--4gbrim.xn----ymcbaaajlc6dj7bxne2c.xn--wgbh1c/"
-            m.add_passthru(url)
-            assert m.passthru_prefixes[0] == clean_url
-
-    run()
-    assert_reset()
-
-
 def test_custom_target(monkeypatch):
     requests_mock = responses.RequestsMock(target="something.else")
     std_mock_mock = responses.std_mock.MagicMock()
@@ -1821,13 +1919,13 @@ def test_custom_target(monkeypatch):
         "http://example.com/other/path/",
     ),
 )
-def test_request_param(url):
+def test_request_param(url):  # type: ignore[misc]
     @responses.activate
     def run():
         params = {"hello": "world", "example": "params"}
         responses.add(
             method=responses.GET,
-            url="{0}?hello=world".format(url),
+            url=f"{url}?hello=world",
             body="test",
             match_querystring=False,
         )
@@ -1864,7 +1962,7 @@ def test_request_param_with_multiple_values_for_the_same_key():
 @pytest.mark.parametrize(
     "url", ("http://example.com", "http://example.com?hello=world")
 )
-def test_assert_call_count(url):
+def test_assert_call_count(url):  # type: ignore[misc]
     @responses.activate
     def run():
         responses.add(responses.GET, url)
@@ -1874,7 +1972,7 @@ def test_assert_call_count(url):
 
         with pytest.raises(AssertionError) as excinfo:
             responses.assert_call_count(url, 2)
-        assert "Expected URL '{0}' to be called 2 times. Called 0 times.".format(
+        assert "Expected URL '{}' to be called 2 times. Called 0 times.".format(
             url
         ) in str(excinfo.value)
 
@@ -1887,7 +1985,7 @@ def test_assert_call_count(url):
         requests.get(url)
         with pytest.raises(AssertionError) as excinfo:
             responses.assert_call_count(url, 3)
-        assert "Expected URL '{0}' to be called 3 times. Called 2 times.".format(
+        assert "Expected URL '{}' to be called 3 times. Called 2 times.".format(
             url
         ) in str(excinfo.value)
 
@@ -1939,6 +2037,70 @@ def test_call_count_without_matcher():
     assert_reset()
 
 
+def test_response_calls_indexing_and_slicing():
+    @responses.activate
+    def run():
+        responses.add(responses.GET, "http://www.example.com")
+        responses.add(responses.GET, "http://www.example.com/1")
+        responses.add(responses.GET, "http://www.example.com/2")
+
+        requests.get("http://www.example.com")
+        requests.get("http://www.example.com/1")
+        requests.get("http://www.example.com/2")
+        requests.get("http://www.example.com/1")
+        requests.get("http://www.example.com")
+
+        # Use of a type hints here ensures mypy knows the difference between index and slice.
+        individual_call: Call = responses.calls[0]
+        call_slice: List[Call] = responses.calls[1:-1]
+
+        assert individual_call.request.url == "http://www.example.com/"
+
+        assert call_slice == [
+            responses.calls[1],
+            responses.calls[2],
+            responses.calls[3],
+        ]
+        assert [c.request.url for c in call_slice] == [
+            "http://www.example.com/1",
+            "http://www.example.com/2",
+            "http://www.example.com/1",
+        ]
+
+    run()
+    assert_reset()
+
+
+def test_response_calls_and_registry_calls_are_equal():
+    @responses.activate
+    def run():
+        rsp1 = responses.add(responses.GET, "http://www.example.com")
+        rsp2 = responses.add(responses.GET, "http://www.example.com/1")
+        rsp3 = responses.add(
+            responses.GET, "http://www.example.com/2"
+        )  # won't be requested
+
+        requests.get("http://www.example.com")
+        requests.get("http://www.example.com/1")
+        requests.get("http://www.example.com")
+
+        assert len(responses.calls) == len(rsp1.calls) + len(rsp2.calls) + len(
+            rsp3.calls
+        )
+        assert rsp1.call_count == 2
+        assert len(rsp1.calls) == 2
+        assert rsp1.calls[0] is responses.calls[0]
+        assert rsp1.calls[1] is responses.calls[2]
+        assert rsp2.call_count == 1
+        assert len(rsp2.calls) == 1
+        assert rsp2.calls[0] is responses.calls[1]
+        assert rsp3.call_count == 0
+        assert len(rsp3.calls) == 0
+
+    run()
+    assert_reset()
+
+
 def test_fail_request_error():
     """
     Validate that exception is raised if request URL/Method/kwargs don't match
@@ -2001,7 +2163,7 @@ def test_fail_request_error():
         ),
     ],
 )
-def test_response_representations(response_params, expected_representation):
+def test_response_representations(response_params, expected_representation):  # type: ignore[misc]
     response = Response(**response_params)
 
     assert str(response) == expected_representation
@@ -2047,7 +2209,7 @@ def test_mocked_responses_list_registered():
         ("http://fizzbuzz/foo", "http://fizzbuzz/foo"),
     ],
 )
-def test_rfc_compliance(url, other_url):
+def test_rfc_compliance(url, other_url):  # type: ignore[misc]
     @responses.activate
     def run():
         responses.add(method=responses.GET, url=url)
@@ -2101,7 +2263,8 @@ def test_responses_reuse():
     assert_reset()
 
 
-async def test_async_calls():
+@pytest.mark.asyncio
+async def test_async_calls():  # type: ignore[misc]
     @responses.activate
     async def run():
         responses.add(
@@ -2155,7 +2318,7 @@ class TestStrictWrapper:
         run_not_strict()
 
     @pytest.mark.parametrize("assert_fired", (True, False, None))
-    def test_nested_decorators(self, assert_fired):
+    def test_nested_decorators(self, assert_fired):  # type: ignore[misc]
         """Validate if assert_all_requests_are_fired is applied from the correct function.
 
         assert_all_requests_are_fired must be applied from the function
@@ -2301,13 +2464,13 @@ class TestUnitTestPatchSetup:
 
     """
 
-    def setup(self):
+    def setup_method(self):
         self.r_mock = responses.RequestsMock(assert_all_requests_are_fired=True)
         self.r_mock.start()
         self.r_mock.get("https://example.com", status=505)
         self.r_mock.put("https://example.com", status=506)
 
-    def teardown(self):
+    def teardown_method(self):
         self.r_mock.stop()
         self.r_mock.reset()
 
@@ -2328,13 +2491,13 @@ class TestUnitTestPatchSetupRaises:
 
     """
 
-    def setup(self):
+    def setup_method(self):
         self.r_mock = responses.RequestsMock()
         self.r_mock.start()
         self.r_mock.get("https://example.com", status=505)
         self.r_mock.put("https://example.com", status=506)
 
-    def teardown(self):
+    def teardown_method(self):
         with pytest.raises(AssertionError) as exc:
             self.r_mock.stop()
         self.r_mock.reset()
@@ -2416,7 +2579,7 @@ class TestMaxRetry:
                 total=total,
                 backoff_factor=0.1,
                 status_forcelist=[500],
-                method_whitelist=["GET", "POST", "PATCH"],
+                allowed_methods=["GET", "POST", "PATCH"],
                 raise_on_status=raise_on_status,
             )
         )
@@ -2448,7 +2611,7 @@ class TestMaxRetry:
         assert_reset()
 
     @pytest.mark.parametrize("raise_on_status", (True, False))
-    def test_max_retries_exceed(self, raise_on_status):
+    def test_max_retries_exceed(self, raise_on_status):  # type: ignore[misc]
         @responses.activate(registry=registries.OrderedRegistry)
         def run():
             url = "https://example.com"
